// loxodrome.frag (copyleft) 2016 Claude Heiland-Allen <claude@mathr.co.uk>
// see: <https://mathr.co.uk/blog/2016-04-11_calendar_2015_loxodrome.html>
#include "Progressive2D.frag"
uniform float time;

// magic tiling constants
const float twopi1 = 0.15915494309189535;
const float sqrtthree2 = 0.8660254037844386;
const float sqrtthree4 = 0.4330127018922193;

// 1 7 magic twist constants
const float cosa = 0.9899494936611666;
const float sina = 0.1414213562373095;
const float hypa = 7.0710678118654755;

// stereographic projection (sphere -> plane)
vec2 stereo(vec3 v) {
  return v.xy / (1.0 - v.z);
}

// stereographic unprojection (plane -> sphere)
vec3 unstereo(vec2 v) {
  vec3 u = vec3(v, 1.0);
  float d = dot(u, u);
  return vec3(2.0 * v, d - 2.0) / d;
}

// animated horizontal rotation
vec3 rotate(vec3 v) {
  float t = time / (30.0 * twopi1);
  return mat3(1.0, 0.0,0.0, 0.0, cos(t), sin(t), 0.0, -sin(t), cos(t)) * vec3(v.x, -v.z, v.y);
}

// exponential map
vec2 logpolar(vec2 v) {
  return vec2(log(length(v)), atan(v.y, v.x)) * twopi1;
}

// twist and scale so the grid lines up
vec2 twist(vec2 v) {
  return mat2(cosa, sina, -sina, cosa) * hypa * v;
}

// generate a repeating grid
vec2 wrap(vec2 v) {
  vec2 u = v;
  u += vec2(0.25, 0.0);
  u -= floor(u);
  u *= 2.0;
  if (u.x >= 1.0) {
    u.y += 0.5;
  }
  u -= floor(u);
  return u.yx;
}

// helper
float distanceSquared(vec2 u, vec2 v) {
  vec2 uv = u - v;
  return dot(uv, uv);
}

// coloured tiling
vec3 hexagon(vec2 v) {
  vec3 colours[5] =
    { vec3(0.0)
    , vec3(1.0)
    , vec3(1.0, 0.5, 0.0)
    , vec3(1.0, 0.5, 0.0)
    , vec3(1.0, 0.5, 0.0)
    };
  vec2 centers[5] =
    { vec2(0.25, sqrtthree4)
    , vec2(0.75, sqrtthree4)
    , vec2(0.5,  sqrtthree2)
    , vec2(1.0, 0.0)
    , vec2(0.0, 0.0)
    };
  vec2 u = v;
  u.y *= sqrtthree2;
  int which = 0;
  float mind = 1.0e10;
  for (int k = 0; k < 5; ++k) {
    float d = distanceSquared(u, centers[k]);
    if (d < mind) {
      which = k;
      mind = d;
    }
  }
  return colours[which];
}

// per pixel rendering pipeline
vec3 color(vec2 v) {
  return hexagon(wrap(twist(logpolar(stereo(rotate(unstereo(v.yx)))))));
}
